@@ -7,6 +7,7 @@ from copy import deepcopy
 from dataclasses import dataclass
 from datetime import date, datetime
 from enum import Enum
+from functools import partial
 from typing import (
     Any,
     Callable,
@@ -342,6 +343,25 @@ def test_extra_allowed():
     assert model.c == 1
 
 
+def test_reassign_instance_method_with_extra_allow():
+    class Model(BaseModel):
+        model_config = ConfigDict(extra='allow')
+        name: str
+
+        def not_extra_func(self) -> str:
+            return f'hello {self.name}'
+
+    def not_extra_func_replacement(self_sub: Model) -> str:
+        return f'hi {self_sub.name}'
+
+    m = Model(name='james')
+    assert m.not_extra_func() == 'hello james'
+
+    m.not_extra_func = partial(not_extra_func_replacement, m)
+    assert m.not_extra_func() == 'hi james'
+    assert 'not_extra_func' in m.__dict__
+
+
 def test_extra_ignored():
     class Model(BaseModel):
         model_config = ConfigDict(extra='ignore')
@@ -495,20 +515,29 @@ def test_frozen_model():
         a: int = 10
 
     m = FrozenModel()
-
     assert m.a == 10
+
     with pytest.raises(ValidationError) as exc_info:
         m.a = 11
     assert exc_info.value.errors(include_url=False) == [
         {'type': 'frozen_instance', 'loc': ('a',), 'msg': 'Instance is frozen', 'input': 11}
     ]
 
+    with pytest.raises(ValidationError) as exc_info:
+        del m.a
+    assert exc_info.value.errors(include_url=False) == [
+        {'type': 'frozen_instance', 'loc': ('a',), 'msg': 'Instance is frozen', 'input': None}
+    ]
+
+    assert m.a == 10
+
 
 def test_frozen_field():
     class FrozenModel(BaseModel):
         a: int = Field(10, frozen=True)
 
     m = FrozenModel()
+    assert m.a == 10
 
     with pytest.raises(ValidationError) as exc_info:
         m.a = 11
@@ -516,6 +545,14 @@ def test_frozen_field():
         {'type': 'frozen_field', 'loc': ('a',), 'msg': 'Field is frozen', 'input': 11}
     ]
 
+    with pytest.raises(ValidationError) as exc_info:
+        del m.a
+    assert exc_info.value.errors(include_url=False) == [
+        {'type': 'frozen_field', 'loc': ('a',), 'msg': 'Field is frozen', 'input': None}
+    ]
+
+    assert m.a == 10
+
 
 def test_not_frozen_are_not_hashable():
     class TestModel(BaseModel):
@@ -656,7 +693,7 @@ def test_validating_assignment_fail(ValidateAssignmentModel):
         {
             'type': 'string_too_short',
             'loc': ('b',),
-            'msg': 'String should have at least 1 characters',
+            'msg': 'String should have at least 1 character',
             'input': '',
             'ctx': {'min_length': 1},
         }
@@ -1474,8 +1511,8 @@ def test_untyped_fields_warning():
     with pytest.raises(
         PydanticUserError,
         match=re.escape(
-            "A non-annotated attribute was detected: `x = 1`. All model fields require a type annotation; "
-            "if `x` is not meant to be a field, you may be able to resolve this error by annotating it "
+            'A non-annotated attribute was detected: `x = 1`. All model fields require a type annotation; '
+            'if `x` is not meant to be a field, you may be able to resolve this error by annotating it '
             "as a `ClassVar` or updating `model_config['ignored_types']`."
         ),
     ):
@@ -2001,6 +2038,38 @@ def test_model_post_init_subclass_private_attrs():
     assert calls == ['C.model_post_init']
 
 
+def test_model_post_init_subclass_setting_private_attrs():
+    """https://github.com/pydantic/pydantic/issues/7091"""
+
+    class Model(BaseModel):
+        _priv1: int = PrivateAttr(91)
+        _priv2: int = PrivateAttr(92)
+
+        def model_post_init(self, __context) -> None:
+            self._priv1 = 100
+
+    class SubModel(Model):
+        _priv3: int = PrivateAttr(93)
+        _priv4: int = PrivateAttr(94)
+        _priv5: int = PrivateAttr()
+        _priv6: int = PrivateAttr()
+
+        def model_post_init(self, __context) -> None:
+            self._priv3 = 200
+            self._priv5 = 300
+            super().model_post_init(__context)
+
+    m = SubModel()
+
+    assert m._priv1 == 100
+    assert m._priv2 == 92
+    assert m._priv3 == 200
+    assert m._priv4 == 94
+    assert m._priv5 == 300
+    with pytest.raises(AttributeError):
+        assert m._priv6 == 94
+
+
 def test_model_post_init_correct_mro():
     """https://github.com/pydantic/pydantic/issues/7293"""
     calls = []
@@ -2976,7 +3045,7 @@ def test_schema_generator_customize_type_constraints_order() -> None:
         {
             'type': 'string_too_long',
             'loc': ('y',),
-            'msg': 'String should have at most 1 characters',
+            'msg': 'String should have at most 1 character',
             'input': ' 1 ',
             'ctx': {'max_length': 1},
         }
